# -*- coding: utf-8 -*-
"""
One turn map matrix elements

@author: gamelina
@date: 04/03/2020
"""

from abc import ABCMeta, abstractmethod
from functools import wraps
import numpy as np

class Element(metaclass=ABCMeta):
    """
    Abstract Element class used for subclass inheritance to define all kinds 
    of objects which intervene in the tracking.
    """

    @abstractmethod
    def track(self, beam):
        """
        Track a beam object through this Element.
        This method needs to be overloaded in each Element subclass.
        """
        raise NotImplementedError
        
    @staticmethod
    def all_bunches(function):
        """
        """
        @wraps(function)
        def wrapper(*args, **kwargs):
            self = args[0]
            beam = args[1]
            if (self.ring.mpi == True):
                rank = self.ring.mpi_init()
                function(self, beam[rank], *args[2:], **kwargs)
            else:
                for bunch in beam:
                    function(self, bunch, *args[2:], **kwargs)
        return wrapper
    
    @staticmethod
    def not_empty(function):
        """
        """
        @wraps(function)
        def wrapper(*args, **kwargs):
            self = args[0]
            beam = args[1]
            if (self.ring.mpi == True):
                rank = self.ring.mpi_init()
                function(self, beam[rank], *args[2:], **kwargs)
            else:
                for bunch in beam.not_empty:
                    function(self, bunch, *args[2:], **kwargs)
        return wrapper

        
class Long_one_turn(Element):
    """
    Longitudinal transform for a single turn.
    """
    
    def __init__(self, ring):
        self.ring = ring
        
    @Element.not_empty
    def track(self, bunch):
        """track"""
        bunch["delta"] -= self.ring.U0/self.ring.E0
        bunch["tau"] -= self.ring.ac*self.ring.T0*bunch["delta"]
        
class SynchrotronRadiation(Element):
    """SyncRad"""
    
    def __init__(self, ring):
        self.ring = ring
        
    @Element.not_empty        
    def track(self, bunch):
        """track"""
        rand = np.random.normal(size=len(bunch))
        bunch["delta"] = ((1 - 2*self.ring.T0/self.ring.tau[2])*bunch["delta"] +
             2*self.ring.sigma_delta*(self.ring.T0/self.ring.tau[2])**0.5*rand)

class RF_cavity(Element):
    """ Perfect RF cavity class for main and harmonic RF cavities."""
    def __init__(self, ring, m, Vc, theta):
        self.ring = ring
        self.m = m # Harmonic number of the cavity
        self.Vc = Vc # Amplitude of Cavity voltage [V]
        self.theta = theta # phase of Cavity voltage
        
    @Element.not_empty    
    def track(self,bunch):
        """Track """
        bunch["delta"] += self.Vc/self.ring.E0*np.cos(self.m*self.ring.omega1*bunch["tau"] + self.theta)
        
        
class Trans_one_turn(Element):
    """
    Transverse transformation for a single turn.
    """
    
    def __init__(self, ring):
        self.ring = ring
        self.alpha = self.ring.mean_optics.alpha
        self.beta = self.ring.mean_optics.beta
        self.gamma = self.ring.mean_optics.gamma
        self.disp = self.ring.mean_optics.disp
        self.dispp = self.ring.mean_optics.dispp
        self.phase_advance = self.ring.tune[0:2]*2*np.pi
    
    @Element.not_empty    
    def track(self, bunch):
        """track"""

        phase_advance_x = self.phase_advance[0]*(1+self.ring.chro[0]*bunch["delta"])
        phase_advance_y = self.phase_advance[1]*(1+self.ring.chro[1]*bunch["delta"])
        matrix = np.zeros((6,6,len(bunch)))
        matrix[0,0,:] = np.cos(phase_advance_x) + self.alpha[0]*np.sin(phase_advance_x)
        matrix[0,1,:] = self.beta[0]*np.cos(phase_advance_x)
        matrix[0,2,:] = self.disp[0]
        matrix[1,0,:] = -1*self.gamma[0]*np.sin(phase_advance_x)
        matrix[1,1,:] = np.cos(phase_advance_x) - self.alpha[0]*np.sin(phase_advance_x)
        matrix[1,2,:] = self.dispp[0]
        matrix[2,2,:] = 1
        
        matrix[3,3,:] = np.cos(phase_advance_y) + self.alpha[1]*np.sin(phase_advance_y)
        matrix[3,4,:] = self.beta[1]*np.cos(phase_advance_y)
        matrix[3,5,:] = self.disp[1]
        matrix[4,3,:] = -1*self.gamma[1]*np.sin(phase_advance_y)
        matrix[4,4,:] = np.cos(phase_advance_y) - self.alpha[1]*np.sin(phase_advance_y)
        matrix[4,5,:] = self.dispp[1]
        matrix[5,5,:] = 1
        
        x = matrix[0,0,:]*bunch["x"] + matrix[0,1,:]*bunch["xp"] + matrix[0,2,:]*bunch["delta"]
        xp = matrix[1,0,:]*bunch["x"] + matrix[1,1,:]*bunch["xp"] + matrix[1,2,:]*bunch["delta"]
        y =  matrix[3,3,:]*bunch["y"] + matrix[3,4,:]*bunch["yp"] + matrix[3,5,:]*bunch["delta"]
        yp = matrix[4,3,:]*bunch["y"] + matrix[4,4,:]*bunch["yp"] + matrix[4,5,:]*bunch["delta"]
        
        bunch["x"] = x
        bunch["xp"] = xp
        bunch["y"] = y
        bunch["yp"] = yp

